# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

bl_info = {
    "name": "Adjust Projection",
    "author": "Witold Jaworski",
    "version": (1, 0, 0),
    "blender": (2, 5, 7),
    "api": 36147,
    "location": "View3D > Properties pane (N-key)",
    "category": "3D View",
    "description": "Adjust view rotation, zoom and pan using numeric controls",
    "warning": "",
    "wiki_url": "",
    "tracker_url": ""
    }

#----- imports
import bpy
from bpy.props import *
from mathutils import *
import math

#----- diagnostic flag 
_DEBUG = True #set it to True, to allow some diagnostic messages in the Blender console

#----- public classes
class CurrentProjectionPanel(bpy.types.Panel):
    bl_idname = "VIEW3D_PT_CurrentProjection"
    bl_label = "Adjust Projection"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    
    @classmethod
    def poll(cls, context):
        space_type = context.area.active_space.rna_type.identifier # quickest way: context.region_data == None
        return ( space_type == 'SpaceView3D')
        
    def draw(self, context):
        region = context.area.active_space.region_3d
        layout = self.layout
        layout.operator_context = 'INVOKE_DEFAULT'
        layout.prop(region,"view_perspective")
        layout.prop(region,"view_distance")
        #layout.column().prop(region,"view_location") #Arrange it in a column!
        layout.operator("view3d.adjust_location")
        #layout.column().prop(region,"view_rotation")
        layout.operator("view3d.adjust_rotation")
         

class View3DAdjustLocation(bpy.types.Operator):
    '''Adjust the numeric values of current view's location in Tools Properties'''
    bl_idname = "view3d.adjust_location"
    bl_label = "View Location"
    bl_description = "Adjust view's location (foucs) in Tool Properties panel"
    bl_options = {'REGISTER', 'UNDO'} #Set this options, if you want to edit in Tools pane the properties of this operator
    
    #--- helper data
    screen_to_world = Matrix() #actual transformation from screen coordinates to world coordinates (set up in invoke())
    view_to_world = Matrix() #actual transformation from view coordinates (screen without perspective) to world coordinates (set up in invoke())
    lo = Vector() #stores the initial view location (set up in invoke())

    #--- operator parameters
    lx = FloatProperty(name="Horizontal adjust.", description="Horizontal movement of the view", step=0.5, precision=3, default = 0)
    ly = FloatProperty(name="Vertical adjust.", description="Vertical movement of the view", step=0.5, precision=3, default = 0)
    lz = FloatProperty(name="Depth adjust.", description="Change depth of the view (important in perspective, only)", step=0.5, precision=3, default = 0)
    #step: increase/decrease when you click on the field left/right arrows. When step = 1, Blender uses 0.01 (it is a percent?)
    #precision: number of displayed digits after decimal dot. 

    #--- Blender interface methods        
    @classmethod
    def poll(cls, context): 
        '''Blender calls it to check, if the operator is available for given context
           When we are not in a 3D View editor, returns False
        '''
        space_type = context.area.active_space.rna_type.identifier
        return (space_type == 'SpaceView3D')
    
    def invoke(self, context, event):
        '''Blender calls it once, when operator is called (by pressing its button on a panel, for example)
           I use it here just to prepare the default values for operator's parameters, displayed in the 
           Tool Properties pane. (They will appear there because of bl_options = {'REGISTE', 'UNDO'})
        '''
        region = context.area.active_space.region_3d
            
        self.screen_to_world = region.perspective_matrix.copy() #perspective matrix describes world-->screen
        self.screen_to_world.invert() #so we have to revert it, to obtain screen-->world transformation
        self.view_to_world = region.view_matrix.copy()
        self.view_to_world.invert()
        self.lo = region.view_location.copy()
        self.lo.resize_4d() #required to match properly perspective matrix
        self.lo *= region.perspective_matrix #reference point - original in the screen coordinates....
        
        self.lz = self.lx = self.ly = 0 #reset the actual offset of the view center:
        
        bpy.ops.view3d.toolshelf() #Show toolshelf
        return {'FINISHED'} 
    
    def execute(self, context):
        '''The most important method of the Operator: does  what it should do.
           It is called from Tool Properties each time, when user change the value of any operator's parameter!
           This way Blender creates impression of "dynamic" response to user's action
        '''
        region = context.area.active_space.region_3d
        
        #calculate current location
        loc = self.lo.copy() + Vector((self.lx, self.ly, 0, 0.0))
        loc *= self.screen_to_world
        if loc[3] != 0.0: loc /= loc[3] #come back from the perspective...
        
        #the depth value we treat in other way (without perspective):
        depth = Vector((0,0,self.lz,1))*self.view_to_world
        
        loc += depth

        loc.resize_3d()
        #update the view:
        region.view_location = loc  

        return {'FINISHED'}

class View3DAdjustRotation(bpy.types.Operator):
    '''Adjust the numeric values of current projection rotation in Tools Properties'''
    bl_idname = "view3d.adjust_rotation"
    bl_label = "View Rotation"
    bl_description = "Adjust view's rotation in Tool Properties panel"
    bl_options = {'REGISTER', 'UNDO'} #Set this options, if you want to edit in Tools pane the properties of this operator
    
    #--- constant data:
    Uo = {'x' : Vector((1,0,0)), 'y' : Vector((0,1,0)), 'z' : Vector((0,0,1)) } #helper: screen coordinate system
    Range = math.pi/2 #rotation range for the operator parameters

    #--- helper data
    base_rotation = Matrix() #initial rotation of the view (set up in invoke())
    
    #--- operator parameters
    ry = FloatProperty(name="Horizontal adjust.", description="Horizontal rotation of the view", 
                       min = -Range, max = Range, default = 0, subtype='ANGLE')
    rx = FloatProperty(name="Vertical adjust.", description="Vertical rotation of the view", 
                       min = -Range, max = Range, default = 0, subtype='ANGLE')
    rz = FloatProperty(name="Self rotation adjust.", description="Rotation around the center of the screen",  
                       min = -Range, max = Range, default = 0, subtype='ANGLE')
    #subtype= 'ANGLE' switches the displays property value in degrees (but it still returns radians, in the program) 
    
    #--- Blender interface methods        
    @classmethod
    def poll(cls, context): 
        space_type = context.area.active_space.rna_type.identifier
        return (space_type == 'SpaceView3D')
    
    def invoke(self, context, event):
        region = context.area.active_space.region_3d
        self.base_rotation = region.view_rotation.to_matrix().to_4x4()
        self.rx = self.ry = self.rz = 0.0 #reset the "working" rotation of the view:
        bpy.ops.view3d.toolshelf() #Show toolshelf
        return {'FINISHED'} 
    
    def execute(self, context):
        region = context.area.active_space.region_3d
        
        #calculate current rotation:
        rotx = self.to_matrix(self.rx,'x')
        roty = self.to_matrix(self.ry,'y')
        rotz = self.to_matrix(self.rz,'z')
        
        rot = rotx*roty*rotz*self.base_rotation
        
        #update the view:
        region.view_rotation = rot.to_quaternion()

        return {'FINISHED'}
    
    #---- Helper functions    
    def to_matrix(self, angle, axis):
        ''' Creates transformation matrix for rotation around selected axis'''
        return Matrix.Rotation(angle, 4, self.Uo[axis])

# Register
def register():
    bpy.utils.register_module(__name__)

def unregister():
    bpy.utils.unregister_module(__name__)

if _DEBUG:
    print("SetView.py: new version loaded!") #just to be sure, that we are testing what we should

if __name__ == '__main__':
    register()
